# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.witness module

Witness command line interface
"""
import argparse
import logging

import falcon
from hio.base import doing
from hio.core import http
from hio.core.tcp import serving as tcpServing
from hio.help import decking

from keri import __version__, kering
from keri import help
from keri.app import directing, habbing, keeping, agenting, indirecting
from keri.core import scheming
from keri.db import basing
from keri.peer import httping, exchanging
from keri.vc import walleting, handling

d = "Runs KERI Agent controller.\n"
d += "Example:\nagent -t 5621\n"
parser = argparse.ArgumentParser(description=d)
parser.set_defaults(handler=lambda args: launch(args))
parser.add_argument('-V', '--version',
                    action='version',
                    version=__version__,
                    help="Prints out version of script runner.")
parser.add_argument('-H', '--http',
                    action='store',
                    default=5620,
                    help="Local port number the HTTP server listens on. Default is 5620.")
parser.add_argument('-T', '--tcp',
                    action='store',
                    default=5621,
                    help="Local port number the HTTP server listens on. Default is 5621.")
parser.add_argument('-a', '--admin-http-port',
                    action='store',
                    default=5623,
                    help="Admin port number the HTTP server listens on. Default is 5623.")
parser.add_argument('-t', '--admin-tcp-port',
                    action='store',
                    default=5624,
                    help="Admin port number the HTTP server listens on. Default is 5624.")
parser.add_argument('-n', '--name',
                    action='store',
                    default="agent",
                    help="Name of controller. Default is agent.")
parser.add_argument('-p', '--pre',
                    action='store',
                    default="",
                    help="Identifier prefix to accept control messages from.")



def launch(args):
    help.ogler.level = logging.INFO
    help.ogler.reopen(name=args.name, temp=True, clear=True)

    logger = help.ogler.getLogger()

    logger.info("\n******* Starting Agent for %s listening: http/%s, tcp/%s "
                ".******\n\n", args.name, args.http, args.tcp)



    runAgent(controller=args.pre, name=args.name,
             httpPort=int(args.http),
             tcp=int(args.tcp),
             adminHttpPort=int(args.admin_http_port),
             adminTcpPort=int(args.admin_tcp_port))

    logger.info("\n******* Ended Agent for %s listening: http/%s, tcp/%s"
                ".******\n\n", args.name, args.http, args.tcp)


def runAgent(controller, name="agent", httpPort=5620, tcp=5621, adminHttpPort=5623, adminTcpPort=5624):
    """
    Setup and run one agent
    """

    ks = keeping.Keeper(name=name, temp=False)  # not opened by default, doer opens
    ksDoer = keeping.KeeperDoer(keeper=ks)  # doer do reopens if not opened and closes
    db = basing.Baser(name=name, temp=False, reload=True)  # not opened by default, doer opens
    dbDoer = basing.BaserDoer(baser=db)  # doer do reopens if not opened and closes

    # setup habitat
    hab = habbing.Habitat(name=name, ks=ks, db=db, temp=False, create=False)
    habDoer = habbing.HabitatDoer(habitat=hab)  # setup doer

    # setup doers
    server = tcpServing.Server(host="", port=tcp)
    tcpServerDoer = tcpServing.ServerDoer(server=server)
    directant = directing.Directant(hab=hab, server=server)

    wallet = walleting.Wallet(hab=hab, name=name)

    jsonSchema = scheming.JSONSchema(resolver=scheming.jsonSchemaCache)
    issueHandler = handling.IssueHandler(wallet=wallet, typ=jsonSchema)
    requestHandler = handling.RequestHandler(wallet=wallet, typ=jsonSchema)
    proofHandler = handling.ProofHandler()
    exchanger = exchanging.Exchanger(hab=hab, handlers=[issueHandler, requestHandler, proofHandler])

    mbx = indirecting.MailboxDirector(hab=hab, exc=exchanger)

    doers = [ksDoer, dbDoer, habDoer, exchanger, directant, tcpServerDoer, mbx]
    doers.extend(adminInterface(controller, hab, proofHandler.proofs, adminHttpPort, adminTcpPort))

    try:
        tock = 0.03125
        doist = doing.Doist(limit=0.0, tock=tock, real=True)
        doist.do(doers=doers)
    except kering.ConfigurationError:
        print(f"prefix for {name} does not exist, incept must be run first", )


def adminInterface(controller, hab, proofs, adminHttpPort=5623, adminTcpPort=5624):
    echoHandler = agenting.EchoHandler()
    rotateHandler = agenting.RotateHandler(hab=hab)
    issueHandler = agenting.IssueCredentialHandler(hab=hab)
    requestHandler = agenting.PresentationRequestHandler(hab=hab)
    handlers = [rotateHandler, issueHandler, requestHandler, echoHandler]

    exchanger = exchanging.Exchanger(hab=hab, controller=controller, handlers=handlers)

    server = tcpServing.Server(host="", port=adminTcpPort)
    tcpServerDoer = tcpServing.ServerDoer(server=server)
    directant = directing.Directant(hab=hab, server=server)

    # app = falcon.App(cors_enable=True)
    app = falcon.App(middleware=falcon.CORSMiddleware(
        allow_origins='*', allow_credentials='*', expose_headers=['cesr-attachment', 'cesr-date', 'content-type']))

    mbx = exchanging.Mailboxer(name=hab.name)
    rep = httping.Respondant(hab=hab, mbx=mbx)
    exnServer = httping.AgentExnServer(exc=exchanger, app=app, rep=rep)
    mbxer = httping.MailboxServer(app=app, hab=hab, mbx=mbx)
    proofHandler = AdminProofHandler(hab=hab, controller=controller, mbx=mbx, proofs=proofs)
    server = http.Server(port=adminHttpPort, app=exnServer.app)
    httpServerDoer = http.ServerDoer(server=server)

    doers = [exchanger, exnServer, tcpServerDoer, directant, httpServerDoer, rep, mbxer, proofHandler]

    return doers


class AdminProofHandler(doing.Doer):
    def __init__(self, hab, controller, mbx, proofs=None, **kwa):
        self.hab = hab
        self.controller = controller
        self.mbx = mbx
        self.proofs = proofs if proofs is not None else decking.Deck()

        super(AdminProofHandler, self).__init__(**kwa)

    def do(self, tymth, tock=0.0, **opts):
        """

        Handle proofs presented externally

        Parameters:
            payload is dict representing the body of a /credential/issue message
            pre is qb64 identifier prefix of sender
            sigers is list of Sigers representing the sigs on the /credential/issue message
            verfers is list of Verfers of the keys used to sign the message

        """

        while True:
            while self.proofs:
                (pre, vc) = self.proofs.popleft()

                pl = dict(
                    pre=pre.qb64,
                    vc=vc
                )

                ser = exchanging.exchange(route="/cmd/presentation/proof", payload=pl, recipient=self.controller)
                msg = self.hab.sanction(ser)

                print("STORING VC PROOF FOR MY CONTROLLER", self.controller)

                self.mbx.storeMsg(self.controller, msg)

                yield

            yield



# -*- encoding: utf-8 -*-
"""
KERI
keri.vdr.eventing module

VC TEL  support
"""


import json
from collections import deque, namedtuple

from orderedset import OrderedSet as oset

from ..core.coring import (MtrDex, Serder, Serials, Versify, Prefixer,
                              Ilks, Seqner, Verfer)
from ..core.eventing import SealEvent, ample, TraitDex, verifySigs, validateSN
from ..db import basing
from ..db.dbing import dgKey, snKey
from ..help import helping
from ..kering import (MissingWitnessSignatureError, Version,
                         MissingAnchorError, ValidationError, OutOfOrderError, LikelyDuplicitousError)
from ..vdr.viring import Registry, nsKey
from .. import help

logger = help.ogler.getLogger()

VCP_LABELS = ["v", "i", "s", "t", "bt", "b", "c"]
VRT_LABELS = ["v", "i", "s", "t", "p", "bt", "b", "ba", "br"]

ISS_LABELS = ["v", "i", "s", "t", "ri", "dt"]
BIS_LABELS = ["v", "i", "s", "t", "ra", "dt"]

REV_LABELS = ["v", "i", "s", "t", "p", "dt"]
BRV_LABELS = ["v", "i", "s", "t", "ra", "p", "dt"]

VcState = namedtuple("VcState", 'issued revoked')

VcStates = VcState(issued='issued', revoked="revoked")


def incept(
        pre,
        toad=None,
        baks=None,
        cnfg=None,
        version=Version,
        kind=Serials.json,
        code=None,
):
    """

    Returns serder of vcp message event
    Utility function to create a Registry inception event

    Parameters:
         pre is issuer identifier prefix qb64
         cnfg is list of strings TraitDex of configuration traits
         toad is int, or str hex of backer threshold
         baks is the initial list of backers prefixes for VCs in the Registry

         version is the API version
         kind is the event type
         code is default code for Prefixer

    """

    vs = Versify(version=version, kind=kind, size=0)
    isn = 0
    ilk = Ilks.vcp

    cnfg = cnfg if cnfg is not None else []

    baks = baks if baks is not None else []
    if TraitDex.NoBackers in cnfg and len(baks) > 0:
        raise ValueError("{} backers specified for NB vcp, 0 allowed".format(len(baks)))

    if len(oset(baks)) != len(baks):
        raise ValueError("Invalid baks = {}, has duplicates.".format(baks))

    if isinstance(toad, str):
        toad = "{:x}".format(toad)
    elif toad is None:
        if not baks:
            toad = 0
        else:  # compute default f and m for len(baks)
            toad = ample(len(baks))

    if baks:
        if toad < 1 or toad > len(baks):  # out of bounds toad
            raise ValueError("Invalid toad = {} for baks = {}".format(toad, baks))
    else:
        if toad != 0:  # invalid toad
            raise ValueError("Invalid toad = {} for baks = {}".format(toad, baks))

    ked = dict(v=vs,  # version string
               i="",  # qb64 prefix
               ii=pre,
               s="{:x}".format(isn),  # hex string no leading zeros lowercase
               t=ilk,
               c=cnfg,
               bt="{:x}".format(toad),  # hex string no leading zeros lowercase
               b=baks  # list of qb64 may be empty
               )

    prefixer = Prefixer(ked=ked, code=code, allows=[MtrDex.Blake3_256])  # Derive AID from ked and code
    ked["i"] = prefixer.qb64  # update pre element in ked with pre qb64

    return Serder(ked=ked)  # return serialized ked


def rotate(
        regk,
        dig,
        sn=1,
        toad=None,
        baks=None,
        cuts=None,
        adds=None,
        version=Version,
        kind=Serials.json,
):
    """

    Returns serder of vrt message event
    Utility function to create a Registry rotation event

    Parameters:
        pre is identifier prefix qb64
        regk is regsitry identifier prefix qb64
        sn is int sequence number
        toad is int or str hex of witness threshold
        baks is list of prior backers prefixes qb64
        cuts is list of witness prefixes to cut qb64
        adds is list of witness prefixes to add qb64

    """

    if sn < 1:
        raise ValueError("Invalid sn = {} for vrt.".format(sn))

    vs = Versify(version=version, kind=kind, size=0)
    ilk = Ilks.vrt

    baks = baks if baks is not None else []
    bakset = oset(baks)
    if len(bakset) != len(baks):
        raise ValueError("Invalid baks = {}, has duplicates.".format(baks))

    cuts = cuts if cuts is not None else []
    cutset = oset(cuts)
    if len(cutset) != len(cuts):
        raise ValueError("Invalid cuts = {}, has duplicates.".format(cuts))

    if (bakset & cutset) != cutset:  # some cuts not in wits
        raise ValueError("Invalid cuts = {}, not all members in baks.".format(cuts))

    adds = adds if adds is not None else []
    addset = oset(adds)
    if len(addset) != len(adds):
        raise ValueError("Invalid adds = {}, has duplicates.".format(adds))

    if cutset & addset:  # non empty intersection
        raise ValueError("Intersecting cuts = {} and  adds = {}.".format(cuts, adds))

    if bakset & addset:  # non empty intersection
        raise ValueError("Intersecting baks = {} and  adds = {}.".format(baks, adds))

    newbakset = (bakset - cutset) | addset

    if len(newbakset) != (len(baks) - len(cuts) + len(adds)):  # redundant?
        raise ValueError("Invalid member combination among baks = {}, cuts ={}, "
                         "and adds = {}.".format(baks, cuts, adds))

    if isinstance(toad, str):
        toad = "{:x}".format(toad)
    elif toad is None:
        if not newbakset:
            toad = 0
        else:  # compute default f and m for len(newbakset)
            toad = ample(len(newbakset))

    if newbakset:
        if toad < 1 or toad > len(newbakset):  # out of bounds toad
            raise ValueError("Invalid toad = {} for resultant wits = {}"
                             "".format(toad, list(newbakset)))
    else:
        if toad != 0:  # invalid toad
            raise ValueError("Invalid toad = {} for resultant wits = {}"
                             "".format(toad, list(newbakset)))

    ked = dict(v=vs,  # version string
               i=regk,  # qb64 prefix
               p=dig,
               s="{:x}".format(sn),  # hex string no leading zeros lowercase
               t=ilk,
               bt="{:x}".format(toad),  # hex string no leading zeros lowercase
               br=cuts,  # list of qb64 may be empty
               ba=adds,  # list of qb64 may be empty
               )

    return Serder(ked=ked)  # return serialized ked


def issue(
        vcdig,
        regk,
        version=Version,
        kind=Serials.json,
):
    """

    Returns serder of iss message event
    Utility function to create a VC issuance event

    Parameters:
        vcdig is hash digest of vc content qb64
        regk is regsitry identifier prefix qb64

    """

    vs = Versify(version=version, kind=kind, size=0)
    ked = dict(v=vs,  # version string
               i=vcdig,  # qb64 prefix
               s="{:x}".format(0),  # hex string no leading zeros lowercase
               t=Ilks.iss,
               ri=regk,
               dt=helping.nowIso8601()
               )

    return Serder(ked=ked)  # return serialized ked


def revoke(
        vcdig,
        regk,
        dig,
        version=Version,
        kind=Serials.json,
):
    """

    Returns serder of rev message event
    Utility function to create a VC revocation vent

    Parameters:
        vcdig is hash digest of vc content qb64
        regk is regsitry identifier prefix qb64
        dig is digest of previous event qb64

    """

    vs = Versify(version=version, kind=kind, size=0)
    isn = 1
    ilk = Ilks.rev

    ked = dict(v=vs,
               i=vcdig,
               s="{:x}".format(isn),  # hex string no leading zeros lowercase
               t=ilk,
               ri=regk,
               p=dig,
               dt=helping.nowIso8601()
               )

    return Serder(ked=ked)  # return serialized ked


def backerIssue(
        vcdig,
        regk,
        regsn,
        regd,
        version=Version,
        kind=Serials.json,
):
    """

    Returns serder of bis message event
    Utility function to create a VC issuance event

    Parameters:
        vcdig is hash digest of vc content qb64
        regk is registry identifier prefix qb64
        regsn is int sequence number of anchoring registry TEL event
        regd is digest qb64 of anchoring registry TEL event

    """

    vs = Versify(version=version, kind=kind, size=0)
    isn = 0
    ilk = Ilks.bis

    seal = SealEvent(regk, regsn, regd)

    ked = dict(v=vs,  # version string
               i=vcdig,  # qb64 prefix
               ii=regk,
               s="{:x}".format(isn),  # hex string no leading zeros lowercase
               t=ilk,
               ra=seal._asdict(),
               dt=helping.nowIso8601(),
               )

    return Serder(ked=ked)  # return serialized ked


def backerRevoke(
        vcdig,
        regk,
        regsn,
        regd,
        dig,
        version=Version,
        kind=Serials.json,
):
    """

    Returns serder of brv message event
    Utility function to create a VC revocation event

    Parameters:
        vcdig is hash digest of vc content qb64
        regk is regsitry identifier prefix qb64
        regsn is int sequence number of anchoring registry TEL event
        regd is digest qb64 of anchoring registry TEL event
        dig is digest of previous event qb64

    """

    vs = Versify(version=version, kind=kind, size=0)
    isn = 1
    ilk = Ilks.brv

    seal = SealEvent(regk, regsn, regd)

    ked = dict(v=vs,
               i=vcdig,
               s="{:x}".format(isn),  # hex string no leading zeros lowercase
               t=ilk,
               p=dig,
               ra=seal._asdict(),
               dt=helping.nowIso8601(),
               )

    return Serder(ked=ked)  # return serialized ked


def query(regk,
          vcid,
          res,
          dt=None,
          dta=None,
          dtb=None,
          version=Version,
          kind=Serials.json):

    """
    Returns serder of query event message.
    Utility function to automate creation of interaction events.

     Parameters:
        pre is identifier prefix qb64
        dig is digest of previous event qb64
        sn is int sequence number
        data is list of dicts of comitted data such as seals
        version is Version instance
        kind is serialization kind
    """
    vs = Versify(version=version, kind=kind, size=0)
    ilk = Ilks.req

    qry = dict(
        i=vcid,
        ri=regk
    )

    if dt is not None:
        qry["dt"] = dt

    if dta is not None:
        qry["dta"] = dt

    if dtb is not None:
        qry["dtb"] = dt


    ked = dict(v=vs,  # version string
               t=ilk,
               r=res,  # resource type for single item request
               q=qry
               )

    return Serder(ked=ked)  # return serialized ked



class Tever:
    """
    Tever is KERI transaction event verifier class
    Only supports current version VERSION

    Has the following public attributes and properties:

    Class Attributes:
        .NoBackers is Boolean
                True means do not allow backers (default to witnesses of controlling KEL)
                False means allow backers (ignore witnesses of controlling KEL)

    Attributes:
        .db is reference to Baser instance that managers the LMDB database
        .reg is regerence to Registry instance that manages VC LMDB database
        .regk is fully qualified base64 identifier prefix of own Registry if any
        .local is Boolean
            True means only process msgs for own events if .regk
            False means only process msgs for not own events if .regk
        .version is version of current event state
        .prefixer is prefixer instance for current event state
        .sn is sequence number int
        .serder is Serder instance of current event with .serder.diger for digest
        .toad is int threshold of accountable duplicity
        .baks is list of qualified qb64 aids for backers
        .cuts is list of qualified qb64 aids for backers cut from prev wits list
        .adds is list of qualified qb64 aids for backers added to prev wits list
        .noBackers is boolean trait True means do not allow backers



    """
    NoBackers = False

    def __init__(self, serder, seqner=None, diger=None, bigers=None, db=None,
                 reger=None, noBackers=None, regk=None, local=False):
        """
        Create incepting tever and state from registry inception serder

        Parameters:
            serder is Serder instance of registry inception event
            seqner (Seqner): issuing event sequence number from controlling KEL.
            diger (Diger): issuing event digest from controlling KEL.
            bigers is list of Siger instances of indexed backer signatures of
                event. Index is offset into baks list of latest est event
            db is Baser instance of lmdb database
            reger is Registry instance of VC lmdb database
            noBackers is boolean True means do not allow backer configuration
            regk is identifier prefix of own or local registry. May not be the
                prefix of this Tever's event. Some restrictions if present
            local is Boolean, True means only process msgs for own controller's
                events if .regk. False means only process msgs for not own events
                if .regk
        """
        self.reger = reger if reger is not None else Registry()
        self.db = db if db is not None else basing.Baser(reopen=True)
        self.version = serder.version
        self.regk = regk
        self.local = True if local else False

        ilk = serder.ked["t"]
        if ilk not in [Ilks.vcp]:
            raise ValidationError("Expected ilk {} got {} for evt: {}".format(Ilks.vcp, ilk, serder))

        self.ilk = ilk
        labels = VCP_LABELS
        for k in labels:
            if k not in serder.ked:
                raise ValidationError("Missing element = {} from {} event for "
                                      "evt = {}.".format(k, ilk, serder.ked))

        self.incept(serder=serder)

        self.config(serder=serder, noBackers=noBackers)

        bigers = self.valAnchorBigs(serder=serder,
                                    seqner=seqner,
                                    diger=diger,
                                    bigers=bigers,
                                    toad=self.toad,
                                    baks=self.baks)

        self.logEvent(pre=self.prefixer.qb64b,
                      sn=0,
                      serder=serder,
                      seqner=seqner,
                      diger=diger,
                      bigers=bigers,
                      baks=self.baks)

    def incept(self, serder):

        ked = serder.ked
        self.pre = ked["ii"]
        self.prefixer = Prefixer(qb64=serder.pre)
        if not self.prefixer.verify(ked=ked, prefixed=True):  # invalid prefix
            raise ValidationError("Invalid prefix = {} for registry inception evt = {}."
                                  .format(self.prefixer.qb64, ked))

        sn = ked["s"]
        self.sn = validateSN(sn, inceptive=True)

        self.cuts = []  # always empty at inception since no prev event
        self.adds = []  # always empty at inception since no prev event
        baks = ked["b"]
        if len(oset(baks)) != len(baks):
            raise ValidationError("Invalid baks = {}, has duplicates for evt = {}."
                                  "".format(baks, ked))
        self.baks = baks

        toad = int(ked["bt"], 16)
        if baks:
            if toad < 1 or toad > len(baks):  # out of bounds toad
                raise ValidationError("Invalid toad = {} for baks = {} for evt = {}."
                                      "".format(toad, baks, ked))
        else:
            if toad != 0:  # invalid toad
                raise ValidationError("Invalid toad = {} for baks = {} for evt = {}."
                                      "".format(toad, baks, ked))
        self.toad = toad
        self.serder = serder

    def config(self, serder, noBackers=None):
        """
        Process cnfg field for configuration traits
        """
        # assign traits
        self.noBackers = (True if (noBackers if noBackers is not None
                                   else self.NoBackers)
                          else False)  # ensure default noBackers is boolean

        cnfg = serder.ked["c"]  # process cnfg for traits
        if TraitDex.NoBackers in cnfg:
            self.noBackers = True

    def update(self, serder, seqner=None, diger=None, bigers=None):
        """
        Process registry non-inception events.
        """

        ked = serder.ked
        ilk = ked["t"]
        sn = ked["s"]

        icp = ilk in (Ilks.iss, Ilks.bis)

        # validate SN for
        sn = validateSN(sn, inceptive=icp)

        if ilk in (Ilks.vrt,):
            if self.noBackers is True:
                raise ValidationError("invalid rotation evt {} against backerless registry {}".
                                      format(ked, self.regk))
            toad, baks, cuts, adds = self.rotate(serder, sn=sn)

            bigers = self.valAnchorBigs(serder=serder,
                                        seqner=seqner,
                                        diger=diger,
                                        bigers=bigers,
                                        toad=toad,
                                        baks=baks)

            self.sn = sn
            self.serder = serder
            self.ilk = ilk
            self.toad = toad
            self.baks = baks
            self.cuts = cuts
            self.adds = adds

            self.logEvent(pre=self.prefixer.qb64b,
                          sn=sn,
                          serder=serder,
                          seqner=seqner,
                          diger=diger,
                          bigers=bigers,
                          baks=self.baks)
            return

        elif ilk in (Ilks.iss, Ilks.bis):
            self.issue(serder, seqner=seqner, diger=diger, sn=sn, bigers=bigers)
        elif ilk in (Ilks.rev, Ilks.brv):
            self.revoke(serder, seqner=seqner, diger=diger, sn=sn, bigers=bigers)
        else:  # unsupported event ilk so discard
            raise ValidationError("Unsupported ilk = {} for evt = {}.".format(ilk, ked))

    def rotate(self, serder, sn):
        """
        Process registry management TEL, non-inception events (vrt)
        """

        ked = serder.ked
        pre = ked["i"]
        dig = ked["p"]

        if serder.pre != self.prefixer.qb64:
            raise ValidationError("Mismatch event aid prefix = {} expecting"
                                  " = {} for evt = {}.".format(ked["i"],
                                                               self.prefixer.qb64,
                                                               ked))
        if not sn == (self.sn + 1):  # sn not in order
            raise ValidationError("Invalid sn = {} expecting = {} for evt "
                                  "= {}.".format(sn, self.sn + 1, ked))

        if not self.serder.compare(dig=dig):  # prior event dig not match
            raise ValidationError("Mismatch event dig = {} with state dig"
                                  " = {} for evt = {}.".format(ked["p"],
                                                               self.serder.diger.qb64,
                                                               ked))

        witset = oset(self.baks)
        cuts = ked["br"]
        cutset = oset(cuts)
        if len(cutset) != len(cuts):
            raise ValidationError("Invalid cuts = {}, has duplicates for evt = "
                                  "{}.".format(cuts, ked))

        if (witset & cutset) != cutset:  # some cuts not in baks
            raise ValidationError("Invalid cuts = {}, not all members in baks"
                                  " for evt = {}.".format(cuts, ked))

        adds = ked["ba"]
        addset = oset(adds)
        if len(addset) != len(adds):
            raise ValidationError("Invalid adds = {}, has duplicates for evt = "
                                  "{}.".format(adds, ked))

        if cutset & addset:  # non empty intersection
            raise ValidationError("Intersecting cuts = {} and  adds = {} for "
                                  "evt = {}.".format(cuts, adds, ked))

        if witset & addset:  # non empty intersection
            raise ValidationError("Intersecting baks = {} and  adds = {} for "
                                  "evt = {}.".format(self.baks, adds, ked))

        baks = list((witset - cutset) | addset)

        if len(baks) != (len(self.baks) - len(cuts) + len(adds)):  # redundant?
            raise ValidationError("Invalid member combination among baks = {}, cuts ={}, "
                                  "and adds = {} for evt = {}.".format(self.baks,
                                                                       cuts,
                                                                       adds,
                                                                       ked))

        toad = int(ked["bt"], 16)
        if baks:
            if toad < 1 or toad > len(baks):  # out of bounds toad
                raise ValidationError("Invalid toad = {} for baks = {} for evt "
                                      "= {}.".format(toad, baks, ked))
        else:
            if toad != 0:  # invalid toad
                raise ValidationError("Invalid toad = {} for baks = {} for evt "
                                      "= {}.".format(toad, baks, ked))

        return toad, baks, cuts, adds

    def issue(self, serder, seqner, diger, sn, bigers=None):
        """
        Process VC TEL issuance events (iss, bis)
        Currently placeholder
        """

        ked = serder.ked
        vcpre = ked["i"]
        ilk = ked["t"]
        vci = nsKey([self.prefixer.qb64, vcpre])

        labels = ISS_LABELS if ilk == Ilks.iss else BIS_LABELS

        for k in labels:
            if k not in ked:
                raise ValidationError("Missing element = {} from {} event for "
                                      "evt = {}.".format(k, ilk, ked))

        if ilk == Ilks.iss:  # simple issue
            if self.noBackers is False:
                raise ValidationError("invalid simple issue evt {} against backer based registry {}".
                                      format(ked, self.regk))

            regi = ked["ri"]
            if regi != self.prefixer.qb64:
                raise ValidationError("Mismatch event regi prefix = {} expecting"
                                      " = {} for evt = {}.".format(regi,
                                                                   self.prefixer.qb64,
                                                                   ked))

            # check if fully anchored
            if not self.verifyAnchor(serder=serder, seqner=seqner, diger=diger):
                self.escrowALEvent(serder=serder)

                raise MissingAnchorError("Failure verify event = {} "
                                         "".format(serder.ked,
                                                   ))

            self.logEvent(pre=vci, sn=sn, serder=serder, seqner=seqner, diger=diger)

        elif ilk == Ilks.bis:  # backer issue
            if self.noBackers is True:
                raise ValidationError("invalid backer issue evt {} against backerless registry {}".
                                      format(ked, self.regk))

            rtoad, baks = self.getBackerState(ked)
            bigers = self.valAnchorBigs(serder=serder,
                                        seqner=seqner,
                                        diger=diger,
                                        bigers=bigers,
                                        toad=rtoad,
                                        baks=baks)

            self.logEvent(pre=vci, sn=sn, serder=serder, seqner=seqner, diger=diger, bigers=bigers)

        else:
            raise ValidationError("Unsupported ilk = {} for evt = {}.".format(ilk, ked))

    def revoke(self, serder, seqner, diger, sn, bigers=None):
        """
        Process VC TEL revocation events (rev, brv)
        Currently placeholder
        """

        ked = serder.ked
        vcpre = ked["i"]
        ilk = ked["t"]

        labels = REV_LABELS if ilk == Ilks.rev else BRV_LABELS

        for k in labels:
            if k not in ked:
                raise ValidationError("Missing element = {} from {} event for "
                                      "evt = {}.".format(k, ilk, ked))

        # have to compare with VC issuance serder
        vci = nsKey([self.prefixer.qb64, vcpre])

        dig = self.reger.getTel(snKey(pre=vci, sn=sn - 1))
        ievt = self.reger.getTvt(dgKey(pre=vci, dig=dig))
        if ievt is None:
            raise ValidationError("revoke without issue... probably have to escrow")

        ievt = bytes(ievt)
        iserder = Serder(raw=ievt)
        if not iserder.compare(dig=ked["p"]):  # prior event dig not match
            raise ValidationError("Mismatch event dig = {} with state dig"
                                  " = {} for evt = {}.".format(ked["p"],
                                                               self.serder.diger.qb64,
                                                               ked))

        if ilk in (Ilks.rev,):  # simple revoke
            if self.noBackers is False:
                raise ValidationError("invalid simple issue evt {} against backer based registry {}".
                                      format(ked, self.regk))

            # check if fully anchored
            if not self.verifyAnchor(serder=serder, seqner=seqner, diger=diger):
                self.escrowALEvent(serder=serder)

                raise MissingAnchorError("Failure verify event = {} "
                                         "".format(serder.ked))

            self.logEvent(pre=vci, sn=sn, serder=serder, seqner=seqner, diger=diger)

        elif ilk in (Ilks.brv,):  # backer revoke
            if self.noBackers is True:
                raise ValidationError("invalid backer issue evt {} against backerless registry {}".
                                      format(ked, self.regk))

            rtoad, baks = self.getBackerState(ked)
            bigers = self.valAnchorBigs(serder=serder,
                                        seqner=seqner,
                                        diger=diger,
                                        bigers=bigers,
                                        toad=rtoad,
                                        baks=baks)

            self.logEvent(pre=vci, sn=sn, serder=serder, seqner=seqner, diger=diger, bigers=bigers)

        else:
            raise ValidationError("Unsupported ilk = {} for evt = {}.".format(ilk, ked))

    def vcState(self, vcpre):
        """
        Calculate state (issued/revoked) of VC from db.
         Returns None if never issued from this Registry

        Parameters:
          vcpre:  the VC identifier
        """
        vci = nsKey([self.prefixer.qb64, vcpre])
        cnt = self.reger.cntTels(vci)
        if cnt == 1:
            return VcStates.issued
        elif cnt == 2:
            return VcStates.revoked

        return None

    def vcSn(self, vcpre):
        """
        Calculates the current seq no of VC from db.
         Returns None if never issued from this Registry

        Parameters:
          vcpre:  the VC identifier

        """
        vci = nsKey([self.prefixer.qb64, vcpre])
        cnt = self.reger.cntTels(vci)

        return None if cnt == 0 else cnt - 1

    def logEvent(self, pre, sn, serder, seqner, diger, bigers=None, baks=None):
        """
        Update associated logs for verified event.
        Update is idempotent. Logs will not write dup at key if already exists.

        Parameters:
            pre (qb64): is event prefix
            sn (int): is event sequence number
            serder (Serder): is Serder instance of current event
            seqner (Seqner): issuing event sequence number from controlling KEL.
            diger (Diger): issuing event digest from controlling KEL.
            bigers (Siger): is optional list of Siger instance of indexed backer sigs
            seqner (Seqner): is optional Seqner instance of cloned first seen ordinal
                If cloned mode then seqner maybe provided (not None)
                When seqner provided then compare fn of dater and database and
                first seen if not match then log and add cue notify problem
            baks (qb64): is optional Dater instance of cloned replay datetime
                If cloned mode then dater maybe provided (not None)
                When dater provided then use dater for first seen datetime
        """

        dig = serder.diger.qb64b
        key = dgKey(pre, dig)
        sealet = seqner.qb64b + diger.qb64b
        self.reger.putAnc(key, sealet)
        if bigers:
            self.reger.putTibs(key, [biger.qb64b for biger in bigers])
        if baks:
            self.reger.delBaks(key)
            self.reger.putBaks(key, [bak.encode("utf-8") for bak in baks])
        self.reger.putTvt(key, serder.raw)
        self.reger.putTel(snKey(pre, sn), dig)
        logger.info("Tever state: %s Added to TEL valid event=\n%s\n",
                    pre, json.dumps(serder.ked, indent=1))

    def valAnchorBigs(self, serder, seqner, diger, bigers, toad, baks):
        """
        Returns double (bigers) where:
        bigers is unique validated signature verified members of inputed bigers

        Validates sigers signatures by validating indexes, verifying signatures, and
            validating threshold sith.
        Validate backer receipts by validating indexes, verifying
            backer signatures and validating toad.
        Backer validation is a function of .regk and .local

        Parameters:
            serder is Serder instance of event
            seqner (Seqner): issuing event sequence number from controlling KEL.
            diger (Diger): issuing event digest from controlling KEL.
            bigers is list of Siger instances of indexed witness signatures.
                Index is offset into wits list of associated witness nontrans pre
                from which public key may be derived.
            toad is int or  str hex of witness threshold
            baks is list of qb64 non-transferable prefixes of backers used to
                derive werfers for bigers

        """

        berfers = [Verfer(qb64=bak) for bak in baks]

        # get unique verified bigers and bindices lists from bigers list
        bigers, bindices = verifySigs(serder=serder, sigers=bigers, verfers=berfers)
        # each biger now has werfer of corresponding wit

        # check if fully anchored
        if not self.verifyAnchor(serder=serder, seqner=seqner, diger=diger):
            self.escrowALEvent(serder=serder, bigers=bigers)

            raise MissingAnchorError("Failure verify event = {} "
                                     "".format(serder.ked))

        # Kevery .process event logic prevents this from seeing event when
        # not local and event pre is own pre
        if ((baks and not self.regk) or  # in promiscuous mode so assume must verify toad
                (baks and not self.local and self.regk and self.regk not in baks)):
            # validate that event is fully witnessed
            if isinstance(toad, str):
                toad = int(toad, 16)
            if toad < 0 or len(baks) < toad:
                raise ValidationError("Invalid toad = {} for wits = {} for evt"
                                      " = {}.".format(toad, baks, serder.ked))

            if len(bindices) < toad:  # not fully witnessed yet
                self.escrowPWEvent(serder=serder, seqner=seqner, diger=diger, bigers=bigers)

                raise MissingWitnessSignatureError("Failure satisfying toad = {} "
                                                   "on witness sigs for {} for evt = {}.".format(toad,
                                                                                                 [siger.qb64 for siger
                                                                                                  in bigers],
                                                                                                 serder.ked))
        return bigers

    def verifyAnchor(self, serder, seqner, diger):
        """
        retrieve event from db using anchor
        get seal from event eserder
        verify pre, sn and dig against serder
        """

        dig = self.db.getKeLast(key=snKey(pre=self.pre, sn=seqner.sn))
        if not dig:
            return False
        else:
            dig = bytes(dig)

        # retrieve event by dig
        raw = self.db.getEvt(key=dgKey(pre=self.pre, dig=dig))
        if not raw:
            return False
        else:
            raw = bytes(raw)

        eserder = Serder(raw=raw)  # deserialize event raw

        if eserder.dig != diger.qb64:
            return False

        seal = eserder.ked["a"]
        if seal is None or len(seal) != 1:
            return False

        seal = seal[0]
        spre = seal["i"]
        ssn = seal["s"]
        sdig = seal["d"]

        if spre == serder.ked["i"] and ssn == serder.ked["s"] \
                and serder.dig == sdig:
            return True

        return False

    def escrowPWEvent(self, serder, seqner, diger, bigers=None):
        """
        Update associated logs for escrow of partially witnessed event

        Parameters:
            serder is Serder instance of  event
            bigers is list of Siger instance of indexed witness sigs
        """
        dgkey = dgKey(serder.preb, serder.digb)
        sealet = seqner.qb64b + diger.qb64b
        self.reger.putAnc(dgkey, sealet)
        self.reger.putTibs(dgkey, [biger.qb64b for biger in bigers])
        self.reger.putTvt(dgkey, serder.raw)
        self.reger.putTwe(snKey(serder.preb, serder.sn), serder.digb)
        logger.info("Tever state: Escrowed partially witnessed "
                    "event = %s\n", serder.ked)

    def escrowALEvent(self, serder, bigers=None):
        """
        Update associated logs for escrow of anchorless event

        Parameters:
            serder is Serder instance of  event
        """
        key = dgKey(serder.preb, serder.digb)
        if bigers:
            self.reger.putTibs(key, [biger.qb64b for biger in bigers])
        self.reger.putTvt(key, serder.raw)
        self.reger.putTae(snKey(serder.preb, serder.sn), serder.digb)
        logger.info("Tever state: Escrowed anchorless event "
                    "event = %s\n", serder.ked)

    def getBackerState(self, ked):
        rega = ked["ra"]
        regi = rega["i"]
        regd = rega["d"]

        if regi != self.prefixer.qb64:
            raise ValidationError("Mismatch event regk prefix = {} expecting"
                                  " = {} for evt = {}.".format(self.regk,
                                                               self.prefixer.qb64,
                                                               ked))

        # load backer list and toad (via event) for specific event in registry from seal in event
        dgkey = dgKey(regi, regd)
        revt = self.reger.getTvt(dgkey)
        if revt is None:
            raise ValidationError("have to escrow this somewhere")

        rserder = Serder(raw=bytes(revt))
        # the backer threshold at this event in mgmt TEL
        rtoad = rserder.ked["bt"]

        baks = [bytes(bak) for bak in self.reger.getBaks(dgkey)]

        return rtoad, baks


class Tevery:
    """
    Tevery (Transaction Event Message Processing Facility)

    Currently placeholder

    Attributes:

    """

    def __init__(self, tevers=None, reger=None, db=None, regk=None, local=False):
        """
        Initialize instance:

        Parameters:
            tevers is dict of Kever instances of key state in db
            reger is Registry instance
            db is Baser instance
            regk is local or own identifier prefix. Some restriction if present
            local is Boolean, True means only process msgs for own events if .pre
                        False means only process msgs for not own events if .pre
        """
        self.tevers = tevers if tevers is not None else dict()
        self.db = db if db is not None else basing.Baser(reopen=True)  # default name = "main"
        self.reger = reger if reger is not None else Registry()
        self.regk = regk  # local prefix for restrictions on local events
        self.local = True if local else False  # local vs nonlocal restrictions
        self.cues = deque()

    def processEvent(self, serder, seqner, diger, wigers=None):
        """
        Process one event serder with attached indexd signatures sigers

        Parameters:
            serder (Serder): event to process
            seqner (Seqner): issuing event sequence number from controlling KEL.
            diger (Diger): issuing event digest from controlling KEL.
            wigers (Siger): is optional list of Siger instances of attached witness indexed sigs

        """
        ked = serder.ked
        try:  # see if code of pre is supported and matches size of pre
            Prefixer(qb64b=serder.preb)
        except Exception as ex:  # if unsupported code or bad size raises error
            raise ValidationError("Invalid pre = {} for evt = {}."
                                  "".format(serder.pre, ked))

        regk = self.registryKey(serder)
        pre = serder.pre
        ked = serder.ked
        sn = ked["s"]
        ilk = ked["t"]

        inceptive = ilk in (Ilks.vcp, Ilks.iss, Ilks.bis)

        # validate SN for
        sn = validateSN(sn, inceptive=inceptive)

        if self.regk:
            if self.local:
                if self.regk != regk:  # nonlocal event when in local mode
                    raise ValueError("Nonlocal event regk={} when local mode for regk={}."
                                     "".format(regk, self.regk))
            else:
                if self.regk == regk:  # local event when not in local mode
                    raise ValueError("Local event regk={} when nonlocal mode."
                                     "".format(regk))

        if regk not in self.tevers:  # first seen for this registry
            if ilk in [Ilks.vcp]:
                # incepting a new registry, Tever create will validate anchor, etc.
                tever = Tever(serder=serder,
                              seqner=seqner,
                              diger=diger,
                              bigers=wigers,
                              reger=self.reger,
                              db=self.db,
                              regk=self.regk,
                              local=self.local)
                self.tevers[regk] = tever
                if not self.regk or self.regk != regk:
                    # witness style backers will need to send receipts so lets queue them up for now
                    self.cues.append(dict(kin="receipt", serder=serder))
            else:
                # out of order, need to escrow
                self.escrowOOEvent(serder=serder, seqner=seqner, diger=diger)
                raise OutOfOrderError("escrowed out of order event {}".format(ked))

        else:
            if ilk in (Ilks.vcp,):
                # we don't have multiple signatures to verify so this
                # is already first seen and then lifely duplicitious
                raise LikelyDuplicitousError("Likely Duplicitous event={}.".format(ked))

            tever = self.tevers[regk]

            if ilk in [Ilks.vrt]:
                sno = tever.sn + 1  # proper sn of new inorder event
            else:
                esn = tever.vcSn(pre)
                sno = 0 if esn is None else esn + 1

            if sn > sno:  # sn later than sno so out of order escrow
                # escrow out-of-order event
                self.escrowOOEvent(serder=serder, seqner=seqner, diger=diger)
                raise OutOfOrderError("Out-of-order event={}.".format(ked))
            elif sn == sno:  # new inorder event
                tever.update(serder=serder, seqner=seqner, diger=diger, bigers=wigers)

                if not self.regk or self.regk != regk:
                    # witness style backers will need to send receipts so lets queue them up for now
                    self.cues.append(dict(kin="receipt", serder=serder))
            else:  # duplicitious
                raise LikelyDuplicitousError("Likely Duplicitous event={} with sn {}.".format(ked, sn))

    def processQuery(self, serder, src=None, sigers=None):
        """
        Process query mode replay message for collective or single element query.
        Assume promiscuous mode for now.

        Parameters:
            serder (Serder) is query message serder
            src (qb64) identifier prefix of event sender
            sigers (list) of Siger instances of attached controller indexed sigs

        """
        ked = serder.ked

        ilk = ked["t"]
        res = ked["r"]
        qry = ked["q"]

        if res == "tels":
            mgmt = qry["ri"]
            vcpre = qry["i"]
            vck = nsKey([mgmt, vcpre])

            cloner = self.reger.clonePreIter(pre=mgmt, fn=0)  # create iterator at 0
            msgs = bytearray()  # outgoing messages
            for msg in cloner:
                msgs.extend(msg)

            cloner = self.reger.clonePreIter(pre=vck, fn=0)  # create iterator at 0
            for msg in cloner:
                msgs.extend(msg)

            self.cues.append(dict(kin="replay", msgs=msgs))
        else:
            raise ValidationError("invalid query message {} for evt = {}".format(ilk, ked))

    @staticmethod
    def registryKey(serder):
        ilk = serder.ked["t"]

        if ilk in (Ilks.vcp, Ilks.vrt):
            return serder.pre
        elif ilk in (Ilks.iss, Ilks.rev):
            return serder.ked["ri"]
        elif ilk in (Ilks.bis, Ilks.brv):
            rega = serder.ked["ra"]
            return rega["i"]
        else:
            raise ValidationError("invalid ilk {} for tevery event = {}".format(ilk, serder.ked))

    def escrowOOEvent(self, serder, seqner, diger):
        key = dgKey(serder.preb, serder.digb)
        self.reger.putTvt(key, serder.raw)
        sealet = seqner.qb64b + diger.qb64b
        self.reger.putAnc(key, sealet)
        self.reger.putOot(snKey(serder.preb, serder.sn), serder.digb)
        logger.info("Tever state: Escrowed anchorless event "
                    "event = %s\n", serder.ked)



# -*- encoding: utf-8 -*-
"""
keri.app.cli.commands module

"""
import argparse

parser = argparse.ArgumentParser(description='Print version of KLI')
parser.set_defaults(handler=lambda args: version(args.hab))


def version(hab):
    print("0.1.0", hab)



# -*- encoding: utf-8 -*-
"""
KERI
keri.vdr.verifying module

VC verifier support
"""

from .. import help
from ..core import parsing, coring
from ..core.coring import Cigar
from ..vdr import eventing
from ..vdr.eventing import VcStates
from ..vdr.viring import Registry

logger = help.ogler.getLogger()


class Verifier:
    """
    Verifier class accepts and validates TEL events.

    """

    def __init__(self, hab, name="test", reger=None, tevers=None, **kwa):
        """
        Initialize Verifier instance

        Parameters:
            hab is Habitat for this verifier's context
            name is user synonym for this verifier
            reger is Registry database instance
            tevers is dict of Tever instances keys by registry identifier
        """
        self.hab = hab
        self.reger = reger if reger is not None else Registry(name=name)
        self.tevers = tevers if tevers is not None else dict()

        self.inited = False

        # save init kwy word arg parameters as ._inits in order to later finish
        # init setup elseqhere after databases are opened if not below
        self._inits = kwa

        if self.hab.inited:
            self.setup()

    def setup(self):
        self.tvy = eventing.Tevery(tevers=self.tevers, reger=self.reger, db=self.hab.db,
                                   regk=None, local=False)
        self.psr = parsing.Parser(framed=True, kvy=self.hab.kvy, tvy=self.tvy)

        self.inited = True

    def verify(self, pre, sidx, regk, vcid, vcdata, vcsig):
        """
        Verifiy the signature and issuance status of a verifiable credential.
        Returns True if the signature is valid

        Parameters:
            pre is qb64 prefix identifier of issuer
            sidx is Int signing key index into issuers keys
            regk is qb64 identifier of the registry
            vcpre is qb64 identifier of VC
            vcdata is the serialized content of the VC
            vcsig is the signature of VC
        """

        state = self.tevers[regk].vcState(vcid)
        if state is None or state is VcStates.revoked:
            return False

        # we don't know about this issuer
        if pre not in self.hab.kevers:
            return False

        kever = self.hab.kevers[pre]
        ksn = kever.state()

        # invalid signature index
        if sidx >= len(ksn.ked["k"]):
            return False

        # assume single signature for now
        verfer = kever.verfers[sidx]

        cigar = Cigar(qb64=vcsig)

        return verfer.verify(sig=cigar.raw, ser=vcdata)

    def query(self, regk, vcid, res, dt=None, dta=None, dtb=None):
        """
        Returns query message for querying for a single element of type res
        """
        kever = self.hab.kever
        serder = eventing.query(regk=regk, vcid=vcid, res=res, dt=dt, dta=dta, dtb=dtb)

        sigers = self.hab.mgr.sign(ser=serder.raw, verfers=kever.verfers)
        msg = bytearray(serder.raw)  # make copy into new bytearray so can be deleted

        msg.extend(coring.Counter(coring.CtrDex.SignerSealCouples, count=1).qb64b)
        msg.extend(self.hab.pre.encode("utf-8"))

        counter = coring.Counter(code=coring.CtrDex.ControllerIdxSigs,
                                 count=len(sigers))
        msg.extend(counter.qb64b)
        for siger in sigers:
            msg.extend(siger.qb64b)

        return msg

    @staticmethod
    def processCuesIter(cues):
        """
        Iterate through cues and yields one or more msgs for each cue.

        Parameters:
            cues is deque of cues

        """
        while cues:  # iteratively process each cue in cues
            cue = cues.popleft()
            cueKin = cue["kin"]  # type or kind of cue

            if cueKin in ("replay",):
                msgs = cue["msgs"]
                yield msgs



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse
import json

from hio.base import doing

from keri.app import habbing
from keri.core.coring import Matter, MtrDex, Diger
from keri.help import helping
from keri.vdr import issuing

parser = argparse.ArgumentParser(description='Issue a verifiable credential')
parser.set_defaults(handler=lambda args: IssueDoer(hab=args.hab,
                                                   recipientIdentifier=args.recipientIdentifier,
                                                   lei=args.lei,
                                                   vcfile=args.vcfile))
parser.add_argument('--recipientIdentifier', '-ri', help='DID Subject ID')
parser.add_argument('--lei', help='Legal Entity Identifier')
parser.add_argument('--vcfile', '-vc', help='output file name')


class IssueDoer(doing.Doer):

    def __init__(self, recipientIdentifier, lei, vcfile, tock=0.0, hab: habbing.Habitat = None, **kwa):
        self.hab = hab
        self.recipientIdentifier = recipientIdentifier
        self.lei = lei
        self.vcfile = vcfile

        super(IssueDoer, self).__init__(tock, **kwa)

    def do(self, tymth, tock=0.0, **opts):
        issuer = issuing.Issuer(hab=self.hab, name=self.hab.name, noBackers=True)
        now = helping.nowIso8601()
        vlei = dict(
            type=[
                "VerifiableCredential",
                "vLEIGLEIFCredential"
            ],
        )
        cred = dict(vlei)
        cred['id'] = "{}".format("#" * Matter.Codes[MtrDex.Blake3_256].fs)
        cred['issuer'] = f"did:keri:{self.hab.pre}"
        cred['issuanceDate'] = now
        cred['credentialSubject'] = dict(
            id=f"did:keri:{self.recipientIdentifier}",
            lei=self.lei
        )

        vcdig = Diger(raw=json.dumps(cred).encode("utf-8"))
        cred['id'] = f"did:keri:{vcdig.qb64}"
        msg = json.dumps(cred).encode("utf-8")

        cigers = self.hab.mgr.sign(ser=msg, verfers=self.hab.kever.verfers, indexed=False)

        cred['proof'] = dict(
            type=[
                "KERISignature2021"
            ],
            created=now,
            jws=cigers[0].qb64,
            verificationMethod=f"did:keri:{self.hab.pre}/{issuer.regk}#0",
            proofPurpose="assertionMethod"
        )

        # tevt, kevt = issuer.issue(vcdig=vcdig.qb64)
        # self.client.tx(kevt)  # send to connected remote
        # logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(kevt))
        # tyme = (yield (self.tock))
        #
        # self.client.tx(tevt)  # send to connected remote
        # logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(tevt))
        # tyme = (yield (self.tock))

        with open(self.vcfile, "w") as f:
            f.write(json.dumps(cred, indent=4))

        print(f'wrote {self.vcfile}')

        return super().do(tymth, tock, **opts)



# -*- encoding: utf-8 -*-
"""
keri.core.scheming module

self-addressing and schema support
"""

import hashlib
import json
from collections import namedtuple

import blake3
import cbor2 as cbor
import jsonschema
import msgpack

from . import coring
from .coring import Matter, MtrDex, Serials
from .. import help
from ..kering import ValidationError, DeserializationError, EmptyMaterialError

Idage = namedtuple("Idage", "dollar at id i")

Ids = Idage(dollar="$id", at="@id", id="id", i="i")

logger = help.ogler.getLogger()


class CacheResolver:
    """
    Sample jsonschema resolver for loading schema $ref references from a local hash.

    """

    def __init__(self, cache=None):
        """
        Create a jsonschema resolver that can be used for loading references to schema remotely.

        Parameters:
            cache (dict) is an optional pre-loaded cache of schema
        """
        self.cache = cache if cache is not None else dict()

    def add(self, key, schema):
        """
        Add schema to cache for resolution

        Parameters:
            key (str) URI to resolve to the schema
            schema (bytes) is bytes of the schema for the URI
        """
        self.cache[key] = schema

    def resolve(self, uri):
        if uri not in self.cache:
            raise ValueError("{} ref not found".format(uri))

        ref = self.cache[uri]
        return ref

    def handler(self, uri):
        """
        Handler provided to jsonschema for cache resolution

        Parameters:
            uri (str) the URI to resolve
        """
        ref = self.resolve(uri)
        schemr = Schemer(raw=ref)
        return schemr.sed

    def resolver(self, scer=b''):
        """
        Returns a jsonschema resolver for returning locally cached schema based on self-addressing
        identifier URIs.

        Parameters:
            scer (bytes) is the source document that is being processed for reference resolution

        """
        return jsonschema.RefResolver("", scer, handlers={"did": self.handler})


jsonSchemaCache = CacheResolver(cache={
    "E7brwlefuH-F_KU_FPWAZR78A3pmSVDlnfJUqnm8Lhr4":
        b'{"$id": "E7brwlefuH-F_KU_FPWAZR78A3pmSVDlnfJUqnm8Lhr4", "$schema": '
        b'"http://json-schema.org/draft-07/schema#", "title": "GLEIF vLEI Credential", "description": "The vLEI '
        b'Credential issued to GLEIF", "type": "object", "properties": {"v": {"type": "string"}, "i": {"type": '
        b'"string"}, "ti": {"type": "string"}, "x": {"description": "schema block", "type": "string"}, '
        b'"d": {"description": "data block", "properties": {"i": {"type": "string"}, "si": {"type": "string"}, '
        b'"issuanceDate": {"format": "date-time", "type": "string"}, "credentialStatus": {"type": "string"}, '
        b'"LEI": {"type": "string"}, "type": {"contains": {"const": "GLEIFvLEICredential"}, "type": "array"}}, '
        b'"additionalProperties": false, "required": ["i", "issuanceDate", "credentialStatus", "LEI", "type"], '
        b'"type": "object"}}, "additionalProperties": false, "required": ["i", "d"]}',
    "EWPMkW-_BU6gh1Y8kizXHchFdmvu_i1wYlYbAC3aJABk":
        b'{"$id": "EWPMkW-_BU6gh1Y8kizXHchFdmvu_i1wYlYbAC3aJABk", "$schema": '
        b'"http://json-schema.org/draft-07/schema#", "title": "Legal Entity Engagement Context Role vLEI Credential", '
        b'"description": "A vLEI Role Credential issued to representatives of a Legal Entity in other than official '
        b'roles but in functional or other context of engagement", "properties": {"v": {"type": "string"}, '
        b'"i": {"type": "string"}, "ti": {"type": "string"}, "x": {"description": "schema block", "type": "string"}, '
        b'"d": {"description": "data block", "properties": {"i": {"type": "string"}, "si": {"type": "string"}, '
        b'"issuanceDate": {"format": "date-time", "type": "string"}, "credentialStatus": {"type": "string"}, '
        b'"LEI": {"type": "string"}, "personLegalName": {"type": "string"}, "engagementContextRole": {"type": '
        b'"string"}, "type": {"contains": {"const": "LegalEntityEngagementContextRolevLEICredential"}, '
        b'"type": "array"}}, "additionalProperties": false, "required": ["i", "issuanceDate", "credentialStatus", '
        b'"LEI", "personLegalName", "engagementContextRole", "type"], "type": "object"}, "s": {"contains": {"type": '
        b'"object"}, "description": "source block", "items": {"properties": {"qualifiedvLEIIssuervLEICredential": {'
        b'"type": "string"}}, "additionalProperties": false, "required": ["qualifiedvLEIIssuervLEICredential"], '
        b'"type": "object"}, "maxItems": 1, "minItems": 1, "type": "array"}}, "additionalProperties": false, '
        b'"required": ["i", "s", "d"], "type": "object"}',
    "EUZ_F1do5sG78zeeA_8CChT5utRpOXQK4GYnv0WGRfuU":
        b'{"$id": "EUZ_F1do5sG78zeeA_8CChT5utRpOXQK4GYnv0WGRfuU", "$schema": '
        b'"http://json-schema.org/draft-07/schema#", "title": "Legal Entity Official Organizational Role vLEI '
        b'Credential", "description": "A vLEI Role Credential issued by a Qualified vLEI issuer to official '
        b'representatives of a Legal Entity", "properties": {"v": {"type": "string"}, "i": {"type": "string"}, '
        b'"ti": {"type": "string"}, "x": {"description": "schema block", "type": "string"}, "d": {"description": '
        b'"data block", "properties": {"i": {"type": "string"}, "si": {"type": "string"}, "issuanceDate": {"format": '
        b'"date-time", "type": "string"}, "credentialStatus": {"type": "string"}, "LEI": {"type": "string"}, '
        b'"personLegalName": {"type": "string"}, "officialRole": {"type": "string"}, "type": {"contains": {"const": '
        b'"LegalEntityOfficialOrganizationalRolevLEICredential"}, "type": "array"}}, "additionalProperties": false, '
        b'"required": ["i", "issuanceDate", "credentialStatus", "LEI", "personLegalName", "officialRole", "type"], '
        b'"type": "object"}, "s": {"contains": {"type": "object"}, "description": "source block", "items": {'
        b'"properties": {"qualifiedvLEIIssuervLEICredential": {"type": "string"}}, "additionalProperties": false, '
        b'"required": ["qualifiedvLEIIssuervLEICredential"], "type": "object"}, "maxItems": 1, "minItems": 1, '
        b'"type": "array"}}, "additionalProperties": false, "required": ["i", "s", "d"], "type": "object"}',
    "E-BRq9StLuC9DxGgiFiy2XND0fFgzyn8cjptlcdvGEFY":
        b'{"$id": "E-BRq9StLuC9DxGgiFiy2XND0fFgzyn8cjptlcdvGEFY", "$schema": '
        b'"http://json-schema.org/draft-07/schema#", "title": "Legal Entity vLEI Credential", "description": "A vLEI '
        b'Credential issued by a Qualified vLEI issuer to a Legal Entity", "properties": {"v": {"type": "string"}, '
        b'"i": {"type": "string"}, "ti": {"type": "string"}, "x": {"description": "schema block", "type": "string"}, '
        b'"d": {"description": "data block", "properties": {"i": {"type": "string"}, "si": {"type": "string"}, '
        b'"issuanceDate": {"format": "date-time", "type": "string"}, "credentialStatus": {"type": "string"}, '
        b'"LEI": {"type": "string"}, "type": {"contains": {"const": "LegalEntityvLEICredential"}, "type": "array"}}, '
        b'"additionalProperties": false, "required": ["i", "issuanceDate", "credentialStatus", "LEI", "type"], '
        b'"type": "object"}, "s": {"contains": {"type": "object"}, "description": "source block", "items": {'
        b'"properties": {"qualifiedvLEIIssuervLEICredential": {"type": "string"}}, "additionalProperties": false, '
        b'"required": ["qualifiedvLEIIssuervLEICredential"], "type": "object"}, "maxItems": 1, "minItems": 1, '
        b'"type": "array"}}, "additionalProperties": false, "required": ["i", "s", "d"], "type": "object"}',
    "E9bX8Do0nb1Eq986HvoJ2iNO00TjC6J_2En8Du9L-hYU":
        b'{"$id": "E9bX8Do0nb1Eq986HvoJ2iNO00TjC6J_2En8Du9L-hYU", "$schema": '
        b'"http://json-schema.org/draft-07/schema#", "title": "Qualified vLEI Issuer Credential", "description": "A '
        b'vLEI Credential issued by GLEIF to Qualified vLEI Issuers which allows the Qualified vLEI Issuers to issue, '
        b'verify and revoke Legal Entity vLEI Credentials and Legal Entity Official Organizational Role vLEI '
        b'Credentials", "properties": {"v": {"type": "string"}, "i": {"type": "string"}, "ti": {"type": "string"}, '
        b'"x": {"description": "schema block", "type": "string"}, "d": {"description": "data block", "properties": {'
        b'"i": {"type": "string"}, "si": {"type": "string"}, "issuanceDate": {"format": "date-time", '
        b'"type": "string"}, "credentialStatus": {"type": "string"}, "LEI": {"type": "string"}, "gracePeriod": {'
        b'"default": 90, "type": "integer"}, "type": {"contains": {"const": "QualifiedvLEIIssuervLEICredential"}, '
        b'"type": "array"}}, "additionalProperties": false, "required": ["i", "issuanceDate", "credentialStatus", '
        b'"LEI", "type"], "type": "object"}}, "additionalProperties": false, "required": ["i", "d"], '
        b'"type": "object"}',
})


class JSONSchema:
    id = Ids.dollar

    def __init__(self, resolver=CacheResolver()):
        self.resolver = resolver

    def resolve(self, uri):
        return self.resolver.resolve(uri)

    def load(self, raw=b'', kind=Serials.json):
        if kind == Serials.json:
            try:
                sed = json.loads(raw.decode("utf-8"))
            except Exception as ex:
                raise DeserializationError("Error deserializing JSON: {} {}"
                                           "".format(raw.decode("utf-8"), ex))

        elif kind == Serials.mgpk:
            try:
                sed = msgpack.loads(raw)
            except Exception as ex:
                raise DeserializationError("Error deserializing MGPK: {} {}"
                                           "".format(raw, ex))

        elif kind == Serials.cbor:
            try:
                sed = cbor.loads(raw)
            except Exception as ex:
                raise DeserializationError("Error deserializing CBOR: {} {}"
                                           "".format(raw, ex))
        else:
            raise ValueError("Invalid serialization kind = {}".format(kind))

        if self.id in sed:
            saider = Saider(qb64=sed[self.id], idder=self.id)
            said = sed[self.id]
            if not saider.verify(sed, prefixed=True):
                raise ValidationError("invalid self-addressing identifier {} instead of {} in schema = {}"
                                      "".format(said, saider.qb64, sed))
        else:
            raise ValidationError("missing ID field {} in schema = {}"
                                  "".format(self.id, sed))

        return sed, kind, saider

    @staticmethod
    def dump(sed, kind=Serials.json):
        raw = coring.dumps(sed, kind)
        return raw

    @staticmethod
    def detect(raw=b''):
        """
        Returns True if content represents JSON Schema by checking
            for $schema;  False otherwise
        """

        try:
            raw.index(b'"$schema"')
        except ValueError:
            return False

        return True

    @staticmethod
    def verify_schema(schema):
        """
        Returns True if the provided schema validates successfully
          as complaint Draft 7 JSON Schema False otherwise

        Parameters:
            schema (dict): is the JSON schema to verify
        """
        try:
            jsonschema.Draft7Validator.check_schema(schema=schema)
        except jsonschema.exceptions.SchemaError:
            return False

        return True

    def verify_json(self, schema=b'', raw=b''):
        """
        Returns True if the JSON passes validation against the
           provided complaint Draft 7 JSON Schema.  Returns False
           if raw is not valid JSON, schema is not valid JSON Schema or
           the validation fails

        Parameters:
              schema (bytes): is the schema use for validation
              raw (bytes): is JSON to validate against the Schema
        """
        try:
            d = json.loads(raw)
            jsonschema.validate(instance=d, schema=schema, resolver=self.resolver.resolver(scer=raw))
        except jsonschema.exceptions.ValidationError as ex:
            logger.error(f'jsonschema.exceptions.ValidationError {ex}')
            return False
        except jsonschema.exceptions.SchemaError as ex:
            logger.error(f'jsonschema.exceptions.SchemaError {ex}')
            return False
        except json.decoder.JSONDecodeError as ex:
            logger.error(f'json.decoder.JSONDecodeError {ex}')
            return False

        return True


class Schemer:
    """
    Schemer is KERI schema serializer-deserializer class
    Verifies self-addressing identifier base on schema type
    Only supports current version VERSION

    Has the following public properties:

    Properties:
        .raw is bytes of serialized event only
        .sed is JSON schema dict
        .kind is Schema kind string value (see namedtuple coring.Serials)
        .saider is Saider instance of self-addressing identifier
        .said  is qb64 digest from .saider

    Hidden Attributes:
          ._raw is bytes of serialized schema only
          ._sed is JSON schema dict
          ._kind is schema kind string value (see namedtuple coring.Serials)
            supported kinds are 'JSONSchema'
          ._code is default code for .saider
          ._saider is Saider instance of digest of .raw


    """

    def __init__(self, raw=b'', sed=None, kind=None, typ=JSONSchema(), code=MtrDex.Blake3_256):
        """
        Deserialize if raw provided
        Serialize if sed provided but not raw
        When serilaizing if kind provided then use kind instead of field in sed

        Parameters:
          raw is bytes of serialized schema
          sed is JSON dict or None
            if None its deserialized from raw
          schemaType is the type of schema
          kind is serialization kind string value or None (see namedtuple coring.Serials)
            supported kinds are 'json', 'cbor', 'msgpack', 'binary'
            if kind is None then its extracted from ked or raw
          code is .saider default digest code

        """

        self._code = code
        self.typ = typ
        if raw:
            self.raw = raw
        elif sed:
            self._kind = kind
            self.sed = sed
        else:
            raise ValueError("Improper initialization need raw or sed.")

        if not self._verify_schema():
            raise ValidationError("invalid kind {} for schema {}"
                                  "".format(self.kind, self.sed))

    def _inhale(self, raw):
        """
        Loads type specific Schema ked and verifies the self-addressing identifier
            of the raw content

        Parameters:
            raw: JSON to load

        """

        sed, kind, saider = self.typ.load(raw=raw)

        return sed, kind, saider

    def _exhale(self, sed, kind=None):
        """
        Dumps type specific Schema JSON and returns the raw bytes, sed
           and schema kind

        Parameters:
            sed: JSON to load
            kind (Schema) tuple of schema type

        """
        saider = Saider(sed=sed, code=self._code, idder=self.typ.id)
        sed[self.typ.id] = saider.qb64
        raw = self.typ.dump(sed)

        return raw, sed, kind, saider

    @property
    def raw(self):
        """ raw property getter """
        return self._raw

    @raw.setter
    def raw(self, raw):
        """ raw property setter """
        sed, kind, saider = self._inhale(raw=raw)
        self._raw = bytes(raw)  # crypto ops require bytes not bytearray
        self._sed = sed
        self._kind = kind
        self._saider = saider

    @property
    def sed(self):
        """ ked property getter"""
        return self._sed

    @sed.setter
    def sed(self, sed):
        """ ked property setter  assumes ._kind """
        raw, sed, kind, saider = self._exhale(sed=sed, kind=self._kind)
        self._raw = raw
        self._kind = kind
        self._sed = sed
        self._saider = saider

    @property
    def kind(self):
        """ kind property getter """
        return self._kind

    @kind.setter
    def kind(self, kind):
        """ kind property setter Assumes ._ked """
        raw, kind, sed = self._exhale(sed=self._sed, kind=kind)
        self._raw = raw
        self._sed = sed
        self._kind = kind
        self._saider = Saider(raw=self._raw, code=self._code)

    @property
    def saider(self):
        """ saider property getter """
        return self._saider

    @property
    def said(self):
        """ said property getter, relies on saider """
        return self.saider.qb64

    def verify(self, raw=b''):
        """
        Returns True if derivation from ked for .code matches .qb64 and
                If prefixed also verifies ked["i"] matches .qb64
                False otherwise

        Parameters:
            raw (bytes): is serialised JSON content to verify against schema
        """

        return self.typ.verify_json(schema=self.sed, raw=raw)

    def _verify_schema(self):
        """
        Returns True if derivation from ked for .code matches .qb64 and
                If prefixed also verifies ked["i"] matches .qb64
                False otherwise

        """

        return self.typ.verify_schema(schema=self.sed)


class Saider(Matter):
    """
    Saider is Matter subclass for self-addressing identifier prefix using
    derivation as determined by code from ked
    """

    Dummy = "#"  # dummy spaceholder char for pre. Must not be a valid Base64 char

    def __init__(self, raw=None, code=None, sed=None, kind=Serials.json, idder=Ids.dollar, **kwa):
        """

        Inherited Parameters:
            raw is bytes of unqualified crypto material usable for crypto operations
            qb64b is bytes of fully qualified crypto material
            qb64 is str or bytes  of fully qualified crypto material
            qb2 is bytes of fully qualified crypto material
            code is str of derivation code

        Parameters:
            sed (dict): optional deserialized JSON for which to create the self addressing
        """

        self.idder = idder
        self._kind = kind
        try:
            # raw is populated
            super(Saider, self).__init__(raw=raw, code=code, **kwa)
        except EmptyMaterialError as ex:
            # No raw, try and calculate code and said

            if not sed or (not code and self.idder not in sed):  # No sed or no code and no id in sed, no luck
                raise ex

            if not code:
                super(Saider, self).__init__(qb64=sed[self.idder], code=code, **kwa)
                code = self.code

            if code == MtrDex.Blake3_256:
                self._derive = self._derive_blake3_256
            elif code == MtrDex.Blake2b_256:
                self._derive = self._derive_blake2b_256
            elif code == MtrDex.Blake2s_256:
                self._derive = self._derive_blake2s_256
            elif code == MtrDex.SHA2_256:
                self._derive = self._derive_sha2_256
            elif code == MtrDex.SHA2_512:
                self._derive = self._derive_sha2_512
            elif code == MtrDex.SHA3_256:
                self._derive = self._derive_sha3_256
            elif code == MtrDex.SHA3_512:
                self._derive = self._derive_sha3_512
            else:
                raise ValueError("Unsupported code = {} for saider.".format(code))

            # use ked and ._derive from code to derive aid prefix and code
            raw, code = self._derive(sed=sed)
            super(Saider, self).__init__(raw=raw, code=code, **kwa)

        if self.code == MtrDex.Blake3_256:
            self._verify = self._verify_blake3_256
        elif self.code == MtrDex.Blake2b_256:
            self._verify = self._verify_blake2b_256
        elif self.code == MtrDex.Blake2s_256:
            self._verify = self._verify_blake2s_256
        elif self.code == MtrDex.SHA3_256:
            self._verify = self._verify_sha3_256
        elif self.code == MtrDex.SHA3_512:
            self._verify = self._verify_sha3_512
        elif self.code == MtrDex.SHA2_256:
            self._verify = self._verify_sha2_256
        elif self.code == MtrDex.SHA2_512:
            self._verify = self._verify_sha2_512
        else:
            raise ValueError("Unsupported code = {} for saider.".format(self.code))

    def derive(self, sed):
        """
        Returns tuple (raw, code) of said as derived from serialized dict sed.
                uses a derivation code specific _derive method

        Parameters:
            sed is json dict

        """

        return self._derive(sed=sed)

    def verify(self, sed, prefixed=False):
        """
        Returns True if derivation from ked for .code matches .qb64 and
                If prefixed also verifies ID value matches .qb64
                False otherwise

        Parameters:
            sed (dict) is json dict
            prefixed (boolean) indicates whether to verify ID value matched .qb64
        """
        try:
            said = self.qb64
            crymat = self._verify(sed=sed)
            if crymat.qb64 != said:
                return False

            idf = self.idder
            if prefixed and sed[idf] != said:
                return False

        except Exception as ex:
            return False

        return True

    def rawify(self, sed):
        if 'v' in sed:
            kind, _, _ = coring.Deversify(sed['v'])
        else:
            kind = self._kind

        raw = coring.dumps(sed, kind)
        return raw

    def _derive_blake3_256(self, sed):
        """
        Returns tuple (raw, code) of basic Blake3 digest (qb64)
            as derived from json dict sed
        """
        sed = dict(sed)  # make copy so don't clobber original sed

        idf = self.idder
        # put in dummy pre to get size correct
        sed[idf] = "{}".format(self.Dummy * Matter.Codes[MtrDex.Blake3_256].fs)

        raw = self.rawify(sed)

        dig = blake3.blake3(raw).digest()
        return dig, MtrDex.Blake3_256

    def _verify_blake3_256(self, sed):
        """
        Returns Matter of typed cryptographic material

        Parameters:
            sed is inception key event dict
        """
        raw, code = self._derive_blake3_256(sed=sed)
        return Matter(raw=raw, code=MtrDex.Blake3_256)

    def _derive_sha3_256(self, sed):
        """
        Returns tuple (raw, code) of basic SHA3 digest (qb64)
            as derived from json dict sed
        """
        sed = dict(sed)  # make copy so don't clobber original sed

        idf = self.idder
        # put in dummy pre to get size correct
        sed[idf] = "{}".format(self.Dummy * Matter.Codes[MtrDex.SHA3_256].fs)
        raw = self.rawify(sed)

        dig = hashlib.sha3_256(raw).digest()
        return dig, MtrDex.SHA3_256

    def _verify_sha3_256(self, sed):
        """
        Returns Matter of typed cryptographic material

        Parameters:
            sed is inception key event dict
        """
        raw, code = self._derive_sha3_256(sed=sed)
        return Matter(raw=raw, code=MtrDex.SHA3_256)

    def _derive_sha3_512(self, sed):
        """
        Returns tuple (raw, code) of basic SHA3 digest (qb64)
            as derived from json dict sed
        """
        sed = dict(sed)  # make copy so don't clobber original sed

        idf = self.idder
        # put in dummy pre to get size correct
        sed[idf] = "{}".format(self.Dummy * Matter.Codes[MtrDex.SHA3_512].fs)
        raw = self.rawify(sed)

        dig = hashlib.sha3_512(raw).digest()
        return dig, MtrDex.SHA3_512

    def _verify_sha3_512(self, sed):
        """
        Returns Matter of typed cryptographic material

        Parameters:
            sed is inception key event dict
        """
        raw, code = self._derive_sha3_512(sed=sed)
        return Matter(raw=raw, code=MtrDex.SHA3_512)

    def _derive_sha2_256(self, sed):
        """
        Returns tuple (raw, code) of basic SHA2 digest (qb64)
            as derived from json dict sed
        """
        sed = dict(sed)  # make copy so don't clobber original sed

        idf = self.idder
        # put in dummy pre to get size correct
        sed[idf] = "{}".format(self.Dummy * Matter.Codes[MtrDex.SHA2_256].fs)
        raw = self.rawify(sed)

        dig = hashlib.sha256(raw).digest()
        return dig, MtrDex.SHA2_256

    def _verify_sha2_256(self, sed):
        """
        Returns Matter of typed cryptographic material

        Parameters:
            sed is schema JSON dict
        """
        raw, code = self._derive_sha2_256(sed=sed)
        return Matter(raw=raw, code=MtrDex.SHA2_256)

    def _derive_sha2_512(self, sed):
        """
        Returns tuple (raw, code) of basic SHA2 digest (qb64)
            as derived from json dict sed
        """
        sed = dict(sed)  # make copy so don't clobber original sed

        idf = self.idder
        # put in dummy pre to get size correct
        sed[idf] = "{}".format(self.Dummy * Matter.Codes[MtrDex.SHA2_512].fs)
        raw = self.rawify(sed)

        dig = hashlib.sha512(raw).digest()
        return dig, MtrDex.SHA2_512

    def _verify_sha2_512(self, sed):
        """
        Returns Matter of typed cryptographic material

        Parameters:
            sed is inception key event dict
            said is Base64 fully qualified default to .qb64
        """
        raw, code = self._derive_sha2_512(sed=sed)
        return Matter(raw=raw, code=MtrDex.SHA2_512)

    def _derive_blake2b_256(self, sed):
        """
        Returns tuple (raw, code) of basic BLAKE2B digest (qb64)
            as derived from json dict sed
        """
        sed = dict(sed)  # make copy so don't clobber original sed

        idf = self.idder
        # put in dummy pre to get size correct
        sed[idf] = "{}".format(self.Dummy * Matter.Codes[MtrDex.Blake2b_256].fs)
        raw = self.rawify(sed)

        dig = hashlib.blake2b(raw).digest()
        return dig, MtrDex.Blake2b_256

    def _verify_blake2b_256(self, sed):
        """
        Returns Matter of typed cryptographic material

        Parameters:
            sed is inception key event dict
        """
        raw, code = self._derive_blake2b_256(sed=sed)
        return Matter(raw=raw, code=MtrDex.Blake2b_256)

    def _derive_blake2s_256(self, sed):
        """
        Returns tuple (raw, code) of basic BLAKE2S digest (qb64)
            as derived from json dict sed
        """
        sed = dict(sed)  # make copy so don't clobber original sed

        idf = self.idder
        # put in dummy pre to get size correct
        sed[idf] = "{}".format(self.Dummy * Matter.Codes[MtrDex.Blake2s_256].fs)
        raw = self.rawify(sed)

        dig = hashlib.blake2s(raw).digest()
        return dig, MtrDex.Blake2s_256

    def _verify_blake2s_256(self, sed):
        """
        Returns Matter of typed cryptographic material

        Parameters:
            sed is inception key event dict
        """
        raw, code = self._derive_blake2s_256(sed=sed)
        return Matter(raw=raw, code=MtrDex.Blake2s_256)
